@@ -1,21 +1,29 @@
 """
-:Date: 2018-08-22
-:Version: 0.1
+:Date: 2022-05-01
+:Version: 0.5
 :Authors: Patrick Erdelt
 
-Demo of TPC-DS in a K8s Cluster.
+    Class for managing an experiment.
+    This is plugged into a cluster object.
+    It collects some configuation objects.
 
-# Compare 4 DBMS on different HW
-# 256 runs
-# no delay
-# Compare result sets
-# 2x each DBMS
-# MemSQL, OmniSci, MonetDB, PostgreSQL, maybe add MySQL, MariaDB, Kinetica?
-# Limit 4 CPUs
+    Two examples included, dealing with TPC-H and TPC-DS tests.
 
-This deals with the TPC-DS tests.
+    Copyright (C) 2020  Patrick Erdelt
+
+    This program is free software: you can redistribute it and/or modify
+    it under the terms of the GNU Affero General Public License as
+    published by the Free Software Foundation, either version 3 of the
+    License, or (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU Affero General Public License for more details.
+
+    You should have received a copy of the GNU Affero General Public License
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
 """
-#from bexhoma import *
 from dbmsbenchmarker import parameter, tools, inspector
 import logging
 import urllib3
@@ -73,11 +81,17 @@ class default():
 			timeout = timeout,
 			singleConnection = True)
 		self.numExperiments = numExperiments
+		self.max_sut = None
 		self.cluster.add_experiment(self)
 		self.appname = self.cluster.appname
 		self.resources = {}
 		self.ddl_parameters = {}
+		self.eval_parameters = {}
 		self.storage = {}
+		self.nodes = {}
+		self.maintaining_parameters = {}
+		self.jobtemplate_maintaining = ""
+		self.jobtemplate_loading = ""
 		#self.connectionmanagement = {}
 		#self.connectionmanagement['numProcesses'] = None
 		#self.connectionmanagement['runsPerConnection'] = None
@@ -86,10 +100,12 @@ class default():
 		self.querymanagement = {}
 		self.workload = {}
 		self.monitoring_active = True
+		self.loading_active = False
 		# k8s:
 		self.namespace = self.cluster.namespace#.config['credentials']['k8s']['namespace']
 		self.configurations = []
 		self.storage_label = ''
+		self.maintaining_active = False
 	def wait(self, sec):
 		print("Waiting "+str(sec)+"s...", end="", flush=True)
 		intervals = int(sec)
@@ -132,8 +148,14 @@ class default():
 		self.resources = {**self.resources, **kwargs}
 	def set_ddl_parameters(self, **kwargs):
 		self.ddl_parameters = kwargs
+	def set_eval_parameters(self, **kwargs):
+		self.eval_parameters = kwargs
 	def set_storage(self, **kwargs):
 		self.storage = kwargs
+	def set_nodes(self, **kwargs):
+		self.nodes = kwargs
+	def set_maintaining_parameters(self, **kwargs):
+		self.maintaining_parameters = kwargs
 	def add_configuration(self, configuration):
 		self.configurations.append(configuration)
 	def __set_queryfile(self, queryfile):
@@ -312,7 +334,7 @@ class default():
 			# include sub directories
 			#cmd['zip_results'] = 'cd /results;zip -r {code}.zip {code}'.format(code=self.code)
 			#fullcommand = 'kubectl exec '+pod_dashboard+' -- bash -c "'+cmd['zip_results'].replace('"','\\"')+'"'
-			self.cluster.executeCTL(command=cmd['zip_results'], pod=pod_dashboard)#self.yamlfolder+deployment)
+			self.cluster.executeCTL(command=cmd['zip_results'], pod=pod_dashboard, container="dashboard")#self.yamlfolder+deployment)
 			#print(fullcommand)
 			#proc = subprocess.Popen(fullcommand, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
 			#stdout, stderr = proc.communicate()
@@ -337,30 +359,59 @@ class default():
 			pods = self.cluster.getPods(component='dashboard')
 			pod_dashboard = pods[0]
 		# copy logs and yamls to result folder
+		print("Copy configuration and logs", end="", flush=True)
 		directory = os.fsencode(self.path)
 		for file in os.listdir(directory):
-			 filename = os.fsdecode(file)
-			 if filename.endswith(".log") or filename.endswith(".yml") or filename.endswith(".error"): 
-				 self.cluster.kubectl('cp '+self.path+"/"+filename+' '+pod_dashboard+':/results/'+str(self.code)+'/'+filename)
+			filename = os.fsdecode(file)
+			if filename.endswith(".log") or filename.endswith(".yml") or filename.endswith(".error"): 
+				self.cluster.kubectl('cp '+self.path+"/"+filename+' '+pod_dashboard+':/results/'+str(self.code)+'/'+filename+' -c dashboard')
+				print(".", end="", flush=True)
+		print("done!")
 		cmd = {}
 		cmd['update_dbmsbenchmarker'] = 'git pull'#/'+str(self.code)
 		#fullcommand = 'kubectl exec '+pod_dashboard+' -- bash -c "'+cmd['update_dbmsbenchmarker'].replace('"','\\"')+'"'
-		self.cluster.executeCTL(command=cmd['update_dbmsbenchmarker'], pod=pod_dashboard)
+		self.cluster.executeCTL(command=cmd['update_dbmsbenchmarker'], pod=pod_dashboard, container="dashboard")
 		#print(fullcommand)
 		#proc = subprocess.Popen(fullcommand, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
 		#stdout, stderr = proc.communicate()
+		print("Join results ", end="", flush=True)
 		cmd['merge_results'] = 'python merge.py -r /results/ -c '+str(self.code)
-		self.cluster.executeCTL(command=cmd['merge_results'], pod=pod_dashboard)
+		self.cluster.executeCTL(command=cmd['merge_results'], pod=pod_dashboard, container="dashboard")
+		print("done!")
 		#fullcommand = 'kubectl exec '+pod_dashboard+' -- bash -c "'+cmd['merge_results'].replace('"','\\"')+'"'
 		#print(fullcommand)
 		#proc = subprocess.Popen(fullcommand, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
 		#stdout, stderr = proc.communicate()
+		print("Build evaluation cube ", end="", flush=True)
 		cmd['evaluate_results'] = 'python benchmark.py read -e yes -r /results/'+str(self.code)
-		self.cluster.executeCTL(command=cmd['evaluate_results'], pod=pod_dashboard)
+		self.cluster.executeCTL(command=cmd['evaluate_results'], pod=pod_dashboard, container="dashboard")
+		print("done!")
 		#fullcommand = 'kubectl exec '+pod_dashboard+' -- bash -c "'+cmd['evaluate_results'].replace('"','\\"')+'"'
 		#print(fullcommand)
 		#proc = subprocess.Popen(fullcommand, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
 		#stdout, stderr = proc.communicate()
+	def stop_maintaining(self):
+		if len(self.configurations) > 0:
+			for config in self.configurations:
+				config.stop_maintaining()
+		else:
+			app = self.cluster.appname
+			component = 'maintaining'
+			configuration = ''
+			jobs = self.cluster.getJobs(app=app, component=component, experiment=self.code, configuration=configuration)
+			for job in jobs:
+				self.cluster.deleteJob(job)
+	def stop_loading(self):
+		if len(self.configurations) > 0:
+			for config in self.configurations:
+				config.stop_loading()
+		else:
+			app = self.cluster.appname
+			component = 'loading'
+			configuration = ''
+			jobs = self.cluster.getJobs(app=app, component=component, experiment=self.code, configuration=configuration)
+			for job in jobs:
+				self.cluster.deleteJob(job)
 	def stop_monitoring(self):
 		if len(self.configurations) > 0:
 			for config in self.configurations:
@@ -375,7 +426,7 @@ class default():
 	def start_monitoring(self):
 		for config in self.configurations:
 			config.start_monitoring()
-	def start_sut(self):
+	def start_sut(self):#, configuration=None):
 		for config in self.configurations:
 			config.start_sut()
 	def stop_sut(self):
@@ -398,33 +449,55 @@ class default():
 	def add_benchmark_list(self, list_clients):
 		for config in self.configurations:
 			config.add_benchmark_list(list_clients)
-	def work_benchmark_list(self, intervals=30):
+	def work_benchmark_list(self, intervals=30, stop=True):
 		do = True
 		while do:
-			time.sleep(intervals)
+			#time.sleep(intervals)
+			self.wait(intervals)
+			# count number of running and pending pods
+			num_pods_running_experiment = len(self.cluster.getPods(app = self.appname, component = 'sut', experiment=self.code, status = 'Running'))
+			num_pods_pending_experiment = len(self.cluster.getPods(app = self.appname, component = 'sut', experiment=self.code, status = 'Pending'))
+			num_pods_running_cluster = len(self.cluster.getPods(app = self.appname, component = 'sut', status = 'Running'))
+			num_pods_pending_cluster = len(self.cluster.getPods(app = self.appname, component = 'sut', status = 'Pending'))
 			for config in self.configurations:
-				# count number of running and pending pods
-				num_pods_running = len(self.cluster.getPods(app = self.appname, component = 'sut', status = 'Running'))
-				num_pods_pending = len(self.cluster.getPods(app = self.appname, component = 'sut', status = 'Pending'))
 				# check if sut is running
 				if not config.sut_is_running():
-					print("{} is not running".format(config.configuration))
+					#print("{} is not running".format(config.configuration))
 					if not config.experiment_done:
 						if not config.sut_is_pending():
-							if self.cluster.max_sut is not None:
-								print("{} running and {} pending pods".format(num_pods_running, num_pods_pending))
-								if num_pods_running+num_pods_pending < self.cluster.max_sut:
+							print("{} is not running yet - ".format(config.configuration))#, end="", flush=True)
+							if self.cluster.max_sut is not None or self.max_sut is not None:
+								we_can_start_new_sut = True
+								if self.max_sut is not None:
+									print("In experiment: {} running and {} pending pods: max is {} pods)".format(num_pods_running_experiment, num_pods_pending_experiment, self.max_sut))#, end="", flush=True)
+									if num_pods_running_experiment+num_pods_pending_experiment >= self.max_sut:
+										print("{} has to wait".format(config.configuration))
+										we_can_start_new_sut = False
+								if self.cluster.max_sut is not None:
+									print("In cluster: {} running and {} pending pods: max is {} pods".format(num_pods_running_cluster, num_pods_pending_cluster, self.cluster.max_sut))#, end="", flush=True)
+									if num_pods_running_cluster+num_pods_pending_cluster >= self.cluster.max_sut:
+										print("{} has to wait".format(config.configuration))
+										we_can_start_new_sut = False
+								if we_can_start_new_sut:
+									print("{} will start now".format(config.configuration))
 									config.start_sut()
-									self.wait(10)
+									num_pods_pending_experiment = num_pods_pending_experiment + 1
+									num_pods_pending_cluster = num_pods_pending_cluster + 1
 							else:
+								print("{} will start now".format(config.configuration))
 								config.start_sut()
-								self.wait(10)
+								num_pods_pending_experiment = num_pods_pending_experiment + 1
+								num_pods_pending_cluster = num_pods_pending_cluster + 1
+								#self.wait(10)
+						else:
+							print("{} is pending".format(config.configuration))
 					continue
 				# check if loading is done
 				config.check_load_data()
 				# start loading
 				if not config.loading_started:
-					print("{} is not loaded".format(config.configuration))
+					if config.sut_is_running():
+						print("{} is not loaded yet".format(config.configuration))
 					if config.monitoring_active and not config.monitoring_is_running():
 						print("{} waits for monitoring".format(config.configuration))
 						if not config.monitoring_is_pending():
@@ -433,7 +506,11 @@ class default():
 					now = datetime.utcnow()
 					if config.loading_after_time is not None:
 						if now >= config.loading_after_time:
-							config.start_loading()
+							if config.loading_active:
+								config.start_loading()
+								config.start_loading_pod(parallelism=config.num_loading)
+							else:
+								config.start_loading()
 						else:
 							print("{} will start loading but not before {}".format(config.configuration, config.loading_after_time.strftime('%Y-%m-%d %H:%M:%S')))
 							continue
@@ -443,13 +520,28 @@ class default():
 							# config demands other delay
 							delay = config.dockertemplate['delay_prepare']
 						config.loading_after_time = now + timedelta(seconds=delay)
-						print("{} will start loading but not before {}".format(config.configuration, config.loading_after_time.strftime('%Y-%m-%d %H:%M:%S')))
+						print("{} will start loading but not before {} (that is in {} secs)".format(config.configuration, config.loading_after_time.strftime('%Y-%m-%d %H:%M:%S'), delay))
 						continue
+				# check if maintaining
+				if config.loading_finished:
+					if config.monitoring_active and not config.monitoring_is_running():
+						print("{} waits for monitoring".format(config.configuration))
+						continue
+					if config.maintaining_active:
+						if not config.maintaining_is_running():
+							print("{} is not maintained yet".format(config.configuration))
+							if not config.maintaining_is_pending():
+								config.start_maintaining(parallelism=config.num_maintaining)
+							else:
+								print("{} has pending maintaining".format(config.configuration))
 				# benchmark if loading is done and monitoring is ready
 				if config.loading_finished:
 					if config.monitoring_active and not config.monitoring_is_running():
 						print("{} waits for monitoring".format(config.configuration))
 						continue
+					if config.maintaining_active and not config.maintaining_is_running():
+						print("{} waits for maintaining".format(config.configuration))
+						continue
 					app = self.cluster.appname
 					component = 'benchmarker'
 					configuration = ''
@@ -473,25 +565,28 @@ class default():
 							config.run_benchmarker_pod(connection=connection, configuration=config.configuration, client=client, parallelism=parallelism)
 						else:
 							# no list element left
-							print("{} can be stopped".format(config.configuration))
-							app = self.cluster.appname
-							component = 'sut'
-							pods = self.cluster.getPods(app, component, self.code, config.configuration)
-							if len(pods) > 0:
-								pod_sut = pods[0]
-								self.cluster.store_pod_log(pod_sut, 'dbms')
-							config.stop_sut()
-							config.numExperimentsDone = config.numExperimentsDone + 1
-							if config.numExperimentsDone < config.numExperiments:
-								print("{} starts again".format(config.configuration))
-								config.benchmark_list = config.benchmark_list_template.copy()
-								# wait for PV to be gone completely
-								self.wait(60)
-								config.reset_sut()
-								config.start_sut()
-								self.wait(10)
+							if stop:
+								print("{} can be stopped".format(config.configuration))
+								app = self.cluster.appname
+								component = 'sut'
+								pods = self.cluster.getPods(app, component, self.code, config.configuration)
+								if len(pods) > 0:
+									pod_sut = pods[0]
+									self.cluster.store_pod_log(pod_sut, 'dbms')
+								config.stop_sut()
+								config.numExperimentsDone = config.numExperimentsDone + 1
+								if config.numExperimentsDone < config.numExperiments:
+									print("{} starts again".format(config.configuration))
+									config.benchmark_list = config.benchmark_list_template.copy()
+									# wait for PV to be gone completely
+									self.wait(60)
+									config.reset_sut()
+									config.start_sut()
+									self.wait(10)
+								else:
+									config.experiment_done = True
 							else:
-								config.experiment_done = True
+								print("{} can be stopped, be we leave it running".format(config.configuration))
 				else:
 					print("{} is loading".format(config.configuration))
 			# all jobs of configuration - benchmarker
@@ -504,7 +599,8 @@ class default():
 			# status per pod
 			for p in pods:
 				status = self.cluster.getPodStatus(p)
-				print(p,status)
+				self.cluster.logger.debug('job-pod {} has status {}'.format(p, status))
+				#print(p,status)
 				if status == 'Succeeded':
 					#if status != 'Running':
 					self.cluster.store_pod_log(p)
@@ -522,7 +618,8 @@ class default():
 			# status per job
 			for job in jobs:
 				success = self.cluster.getJobStatus(job)
-				print(job, success)
+				self.cluster.logger.debug('job {} has success status {}'.format(job, success))
+				#print(job, success)
 				if success:
 					self.cluster.deleteJob(job)
 			if len(pods) == 0 and len(jobs) == 0:
@@ -530,13 +627,13 @@ class default():
 				for config in self.configurations:
 					#if config.sut_is_pending() or config.loading_started or len(config.benchmark_list) > 0:
 					if config.sut_is_pending():
-						print("{} pending".format(config.configuration))
+						self.cluster.logger.debug("{} pending".format(config.configuration))
 						do = True
 					if not config.loading_started:
-						print("{} not loaded".format(config.configuration))
+						self.cluster.logger.debug("{} not loaded".format(config.configuration))
 						do = True
 					if len(config.benchmark_list) > 0:
-						print("{} still benchmarks to run".format(config.configuration))
+						self.cluster.logger.debug("{} still benchmarks to run".format(config.configuration))
 						do = True
 	def benchmark_list(self, list_clients):
 		for i, parallelism in enumerate(list_clients):
@@ -588,11 +685,11 @@ class default():
 						self.cluster.deleteJob(job)
 				if len(pods) == 0 and len(jobs) == 0:
 					break
-	def stop_benchmarker(self):
+	def stop_benchmarker(self, configuration=''):
 		# all jobs of configuration - benchmarker
+		self.cluster.logger.debug("experiment.stop_benchmarker({})".format(configuration))
 		app = self.appname
 		component = 'benchmarker'
-		configuration = ''
 		jobs = self.cluster.getJobs(app, component, self.code, configuration)
 		# status per job
 		for job in jobs:
@@ -663,3 +760,80 @@ class tpch(default):
 	def set_queries_profiling(self):
 		self.set_queryfile('queries-tpch-profiling.config')
 
+class iot(default):
+	def __init__(self,
+			cluster,
+			code=None,
+			queryfile = 'queries-iot.config',
+			SF = '1',
+			numExperiments = 1,
+			timeout = 7200,
+			detached=False):
+		default.__init__(self, cluster, code, numExperiments, timeout, detached)
+		self.set_experiment(volume='iot')
+		self.set_experiment(script='SF'+str(SF)+'-index')
+		self.cluster.set_configfolder('experiments/iot')
+		parameter.defaultParameters = {'SF': str(SF)}
+		self.set_queryfile(queryfile)
+		self.set_workload(
+			name = 'IoT Queries SF='+str(SF),
+			info = 'This experiment performs some IoT inspired queries.'
+			)
+		self.storage_label = 'tpch-'+str(SF)
+		self.maintaining_active = True
+	def set_queries_full(self):
+		self.set_queryfile('queries-iot.config')
+	def set_queries_profiling(self):
+		self.set_queryfile('queries-iot-profiling.config')
+	def set_querymanagement_maintaining(self,
+			numRun=128,
+			delay=5,
+			datatransfer=False):
+		self.set_querymanagement(
+			numWarmup = 0,
+			numCooldown = 0,
+			numRun = numRun,
+			delay = delay,
+			)
+		#self.monitoring_active = True
+		self.maintaining_active = True
+
+class tsbs(default):
+	def __init__(self,
+			cluster,
+			code=None,
+			queryfile = 'queries-tsbs.config',
+			SF = '1',
+			numExperiments = 1,
+			timeout = 7200,
+			detached=False):
+		default.__init__(self, cluster, code, numExperiments, timeout, detached)
+		self.set_experiment(volume='tsbs')
+		self.set_experiment(script='SF'+str(SF)+'-index')
+		self.cluster.set_configfolder('experiments/tsbs')
+		parameter.defaultParameters = {'SF': str(SF)}
+		self.set_queryfile(queryfile)
+		self.set_workload(
+			name = 'TSBS Queries SF='+str(SF),
+			info = 'This experiment performs some TSBS inspired queries.'
+			)
+		self.storage_label = 'tsbs-'+str(SF)
+		self.maintaining_active = True
+		self.jobtemplate_maintaining = "jobtemplate-maintaining-tsbs.yml"
+	def set_queries_full(self):
+		self.set_queryfile('queries-tsbs.config')
+	def set_queries_profiling(self):
+		self.set_queryfile('queries-tsbs-profiling.config')
+	def set_querymanagement_maintaining(self,
+			numRun=128,
+			delay=5,
+			datatransfer=False):
+		self.set_querymanagement(
+			numWarmup = 0,
+			numCooldown = 0,
+			numRun = numRun,
+			delay = delay,
+			)
+		#self.monitoring_active = True
+		self.maintaining_active = True
+
